from transformers import MBartForConditionalGeneration, VisionEncoderDecoderModel
import torch.nn as nn
from torch.nn import functional as F
import py_vncorenlp
import sentencepiece
import math
import torch

class VNTrOCR(nn.Module):
    def __init__(self):
        super().__init__(config=None)
        self.ViTEncoder = VisionEncoderDecoderModel.from_pretrained("microsoft/trocr-small-handwritten").encoder
        self.adaptive_layer = nn.Linear(384, 768)
        bartpho = MBartForConditionalGeneration.from_pretrained("vinai/bartpho-syllable-base")
        self.BartDecoder = bartpho.model.decoder
        self.fc_out = bartpho.lm_head
        self.pad_token_id = 1

    def forward(self, pixel_values, tgt_inputs, att_mask=None):
        """
        `pixel_values`: B,C,H,W
        `tgt_inputs`: B, tgt_len
        `att_mask`: B, tgt_len
        """
        encoder_output = self.ViTEncoder(pixel_values)[0] # B, src_len, 384
        encoder_output = self.adaptive_layer(encoder_output) # B, src_len 768

        max_input_len = att_mask.sum(dim=-1).max().item()

        decoder_output = self.BartDecoder(input_ids=tgt_inputs[:, :max_input_len-1],
                                    attention_mask=att_mask[:, :max_input_len-1],
                                    encoder_hidden_states=encoder_output) # B, tgt_len, 768
        logits = self.fc_out(decoder_output[0]) # B, tgt_len, vocab_size
        return F.cross_entropy(input=logits.transpose(1,2),
                                        target=tgt_inputs[:, 1:max_input_len],
                                        ignore_index=self.pad_token_id, label_smoothing=0.1)
    
    def forward_logit(self, pixel_values, tgt_inputs, att_mask=None):
        encoder_output = self.ViTEncoder(pixel_values)[0] # B, src_len, 384
        encoder_output = self.adaptive_layer(encoder_output) # B, src_len 768

        max_input_len = att_mask.sum(dim=-1).max().item()

        decoder_output = self.BartDecoder(input_ids=tgt_inputs[:, :max_input_len-1],
                                    attention_mask=att_mask[:, :max_input_len-1],
                                    encoder_hidden_states=encoder_output) # B, tgt_len, 768
        logits = self.fc_out(decoder_output[0]) # B, tgt_len, vocab_size
        return logits, max_input_len
    
    def forward_encode(self, pixel_values):
        encoder_output = self.ViTEncoder(pixel_values)[0] # B, src_len, 384
        encoder_output = self.adaptive_layer(encoder_output) # B, src_len 768
        return encoder_output
    
    def forward_decode(self, prev_ids, encoder_output):
        decoder_output = self.BartDecoder(input_ids=prev_ids,
                                          encoder_hidden_states=encoder_output) # B, tgt_len, 768
        logits = self.fc_out(decoder_output[0]) # B, tgt_len, vocab_size
        return logits

class Rethinking(nn.Module):
    def __init__(self, config, masked=True):
        super().__init__()
        self.vocab_size = config.vocab_size
        self.rank = config.rank
        self.block_size = config.max_seq_len
        self.c_attn = nn.Linear(self.vocab_size, self.rank*3, bias=config.bias)
        self.c_proj = nn.Linear(self.rank, self.vocab_size, bias=config.bias)
        self.attn_dropout = nn.Dropout(0)
        self.resid_dropout = nn.Dropout(0.1)
        self.gelu = nn.GELU()
        self.masked = masked


    def forward(self, x):
        B, T, C = x.shape                                               #batch_size, block_size, n_embd
        q, k, v = self.c_attn(x).split(self.rank, dim=2)

        # q = q.view(B, T, self.n_heads, C//self.n_heads).transpose(1,2)  # (B, n_heads, T, h_size)
        # k = k.view(B, T, self.n_heads, C//self.n_heads).transpose(1,2)
        # v = v.view(B, T, self.n_heads, C//self.n_heads).transpose(1,2)

        att = q@k.transpose(-2,-1)*(1/math.sqrt(k.size(-1)))            # (B, T, T)
        if self.masked:
            mask = torch.tril(torch.ones((T, T))).view(1, T, T).to(x.device)
            att = att.masked_fill_(mask == 0, -float('inf'))

        att = F.softmax(att, dim=-1)
        att = self.attn_dropout(att)
        y = att@v                                                       #(B, T, n_embd)
        # y = y.transpose(1,2).contiguous().view(B,T,C)
        y = self.resid_dropout(self.c_proj(y))
        return self.gelu(y)


class AdapterVNTrOCR(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.basemodel = VNTrOCR()
        self.rethinking = Rethinking(config)
        # self.squeeze = nn.Linear(config.vocab_size, config.rank)
        # self.dropout = nn.Dropout(0.1)
        # self.unsqueeze = nn.Linear(config.rank, config.vocab_size)
    
    def forward(self, pixel_values, tgt_inputs, att_mask=None):
        logits, max_input_len = self.basemodel.forward_logit(pixel_values, tgt_inputs, att_mask)
        logits = logits + self.rethinking(logits)
        # logits = self.dropout(self.squeeze(logits))
        # logits = F.relu(logits)
        # logits = self.unsqueeze(logits)
        return  F.cross_entropy(input=logits.transpose(1,2),
                                target=tgt_inputs[:, 1:max_input_len],
                                ignore_index=self.basemodel.pad_token_id, label_smoothing=0.1)
    
    def forward_adapt(self, logits):
        logits = logits + self.rethinking(logits)
        # logits = self.dropout(self.squeeze(logits))
        # logits = self.unsqueeze(logits)
        return logits




from models import *
from utils import *
from dataset import *
from torch.utils.data import DataLoader
import pandas as pd
from tqdm import tqdm
from evaluate import load
import wandb
import math
from jit import OCRModel

device="cuda:1"

if __name__ == '__main__':
    #define model
    torch.set_default_dtype(torch.float32)
    model = AdapterVGGTransformer(config)
    model.load_state_dict(torch.load("/mnt/disk4/VN_HTR/VN_HTR/source/weights/finetune_adapter_vgg_transformer_v3.pt", map_location=device))
    # model = VGGTransformer(config)
    # model.load_state_dict(torch.load("/mnt/disk4/VN_HTR/VN_HTR/source/weights/cp_vgg_transformer_finetune_v4.pt", map_location=device))
    model.to(device)
    # model = OCRModel(device=device)
    # print(model.encoder)
    # print(model.decoder)

    #data loader
    # anot = pd.read_csv("concated_anot.csv").sample(frac=1).reset_index(drop=True)
    
    anot = pd.read_csv("test_anot.csv")

    valdataset = CustomDataset(anot, config)
    valloader = DataLoader(dataset=valdataset, batch_size=128, shuffle=False, generator=torch.Generator(device="cpu"), num_workers=8)

    #utils
    cer = load("cer")
    wer = load("wer")
    cer_score = 0
    wer_score = 0
    tkz = Tokenizer(config)

    print("Starting ....")

    with torch.no_grad():
        # model.eval()
        predictions, references = [], []
        torch.cuda.empty_cache()
        for batch in tqdm(valloader):
            img = batch["image"].to(device)
            tokens = batch["tokens"][:,1:]
            max_width = batch["width"].max().item()

            t_preds = infer_transformer_base(img[:, :, :, :max_width], model.to(device), backbone="vgg", adapter=True)
            # t_preds = model.predict(img[:, :, :, :max_width])

            chars_true = [tkz.reverse_tokens(token) for token in tokens]
            chars_pred = [tkz.reverse_tokens(t_pred) for t_pred in t_preds]

            predictions += chars_pred
            references += chars_true

        cer_score += cer.compute(predictions=predictions, references=references, concatenate_texts=True)
        wer_score += wer.compute(predictions=predictions, references=references, concatenate_texts=True)
        print(f'Character Error Rate: {cer_score}')
        print(f'Word Error Rate: {wer_score}')


import torch.nn as nn
import torch
# from config import predictor, device
# from submodules import *
# import math
# from models import *
# from utils import *
# from dataset import *
# from torch.utils.data import DataLoader
# import pandas as pd
import numpy as np
from torch.nn import functional as F

class OCRModel():
    def __init__(self, adapter=False, device="cuda:2"):
        if adapter:
            self.encoder = torch.jit.load("vnhtr/weights/vta_encoder.pt", map_location=device)#.to(device)
            self.decoder = torch.jit.load("vnhtr/weights/vta_decoder.pt", map_location=device)#.to(device)
        else:
            self.encoder = torch.jit.load("vnhtr/weights/vt_encoder.pt", map_location=device)#.to(device)
            self.decoder = torch.jit.load("vnhtr/weights/vt_decoder.pt", map_location=device)#.to(device)
        print("loaded model")
        self.encoder.eval()
        self.decoder.eval()
        self.device = device
    
    def predict(self, img, max_seq_length=128, sos_token=1, eos_token=2):
        with torch.no_grad():
            with torch.jit.optimized_execution(True):
                mem = self.encoder(img.to(self.device))
                translated_sentence = [[sos_token]*len(img)]
                max_length = 0

                while max_length <= max_seq_length and not all(np.any(np.asarray(translated_sentence).T==eos_token, axis=1)):

                    tgt_inp = torch.LongTensor(translated_sentence).to(self.device)
                    # print(self.gen_nopeek_mask(tgt_inp.shape[0]).to(self.device).device)
                    output = self.decoder(tgt_inp, mem, self.gen_nopeek_mask(tgt_inp.shape[0]).to(self.device))
                    output = F.softmax(output, dim=-1)
                    output = output.to('cpu')

                    values, indices  = torch.topk(output, 5)
                    
                    indices = indices[:, -1, 0]
                    indices = indices.tolist()

                    translated_sentence.append(indices)   
                    max_length += 1

                    del output

                translated_sentence = np.asarray(translated_sentence).T
            
                return translated_sentence
    
    def predict_topk(self, img, max_seq_length=128, sos_token=1, eos_token=2):
        conflict=[]
        with torch.no_grad():
            mem = self.encoder(img.to(self.device))
            translated_sentence = [[sos_token]*len(img)]

            probs = []
            tokens = []

            max_length = 0

            while max_length <= max_seq_length and not all(np.any(np.asarray(translated_sentence).T==eos_token, axis=1)):

                tgt_inp = torch.LongTensor(translated_sentence).to(self.device)
                
                output = self.decoder(tgt_inp, mem, self.gen_nopeek_mask(tgt_inp.shape[0]).to(self.device))
                output = F.softmax(output, dim=-1)
                output = output.to('cpu')

                values, indices  = torch.topk(output, 233)

                probs.append(values[0, -1].cpu().numpy())
                tokens.append(indices[0, -1].cpu().numpy())
            
                indices = indices[:, -1, 0]
                indices = indices.tolist()

                translated_sentence.append(indices)   
                max_length += 1

                conflict.append(output[:,-1,:].cpu().numpy())

                del output

            translated_sentence = np.asarray(translated_sentence).T
        return translated_sentence, np.array(tokens).T, np.array(probs).T, np.array(conflict)
    
    def gen_nopeek_mask(self, length):
        mask = (torch.triu(torch.ones(length, length)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask

class EncoderVGGTrans(nn.Module):
    def __init__(self, model):
        super().__init__()
        # self.backbone = model.basemodel.backbone
        # self.d_model = model.basemodel.transformerLM.d_model
        # self.encoder = model.basemodel.transformerLM.transformer.encoder
        # self.pos_enc = model.basemodel.transformerLM.pos_enc
        self.backbone = model.backbone
        self.d_model = model.transformerLM.d_model
        self.encoder = model.transformerLM.transformer.encoder
        self.pos_enc = model.transformerLM.pos_enc

    def forward(self, img):
        src = self.backbone(img)
        src = self.pos_enc(src*math.sqrt(self.d_model))
        memory = self.encoder(src)
        return memory

class DecoderVGGTrans(nn.Module):
    def __init__(self, model):
        super().__init__()
        # self.decoder = model.basemodel.transformerLM.transformer.decoder
        # self.d_model = model.basemodel.transformerLM.d_model
        # self.embed_tgt = model.basemodel.transformerLM.embed_tgt
        # self.pos_enc = model.basemodel.transformerLM.pos_enc
        # self.fc = model.basemodel.transformerLM.fc
        # self.rethinking = model.rethinking
        # self.fc_out = model.fc_out
        self.decoder = model.transformerLM.transformer.decoder
        self.d_model = model.transformerLM.d_model
        self.embed_tgt = model.transformerLM.embed_tgt
        self.pos_enc = model.transformerLM.pos_enc
        self.fc = model.transformerLM.fc

    def forward(self, tgt, memory, tgt_mask):
        # tgt_mask = self.gen_nopeek_mask(tgt.shape[0]).to(tgt.device)
        tgt = self.pos_enc(self.embed_tgt(tgt) * math.sqrt(self.d_model))
        de_out = self.decoder(tgt, memory, tgt_mask=tgt_mask)
        de_out = self.fc(de_out.transpose(0, 1))

        # de_out = de_out + self.rethinking(de_out, tgt_mask)
        # return self.fc_out(de_out)
        return de_out

device="cuda:3"

def gen_nopeek_mask(length):
    mask = (torch.triu(torch.ones(length, length)) == 1).transpose(0, 1)
    mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
    return mask

if __name__ == "__main__":
    torch.set_default_dtype(torch.float32)
    model = VGGTransformer(config)
    model.load_state_dict(torch.load("/mnt/disk4/VN_HTR/VN_HTR/VGGTransformer/weights/cp_vgg_transformer_finetune_v4.pt", map_location=device))
    model.to(device)

    #data loader
    # anot = pd.read_csv("concated_anot.csv").sample(frac=1).reset_index(drop=True)
    anot = pd.read_csv("test_anot.csv").sample(frac=1, random_state=0).reset_index(drop=True)

    valdataset = CustomDataset(anot, config)
    valloader = DataLoader(dataset=valdataset, batch_size=2, shuffle=False, generator=torch.Generator(device="cpu"))

    for batch in valloader:
        img = batch["image"].to(device)
        in_tokens = batch["tokens"][:,:-1].to(device)
        max_width = batch["width"].max().item()

        encoder = EncoderVGGTrans(model)
        encoder.eval()
        trace_encoder = torch.jit.trace(encoder, img[:, :, :, :max_width])
        trace_encoder.save("/mnt/disk4/VN_HTR/VN_HTR/vnhtr/weights/vt_encoder.pt")
        memory = encoder(img[:, :, :, :max_width])#.transpose(1, 0)
        # print(memory.shape)
        # print(in_tokens[:, :5].shape)

        decoder = DecoderVGGTrans(model)
        decoder.eval()
        trace_decoder = torch.jit.trace(decoder, (in_tokens[:, :3].transpose(0,1).to(device), memory, gen_nopeek_mask(in_tokens[:, :3].shape[1]).to(device)))
        trace_decoder.save("/mnt/disk4/VN_HTR/VN_HTR/vnhtr/weights/vt_decoder.pt")
        
        break




import cv2
import numpy as np
import random
from tqdm import tqdm
import pandas as pd

def add_line(img):

    h, w, c = img.shape
    line_mask = np.ones((h,w,c), dtype=np.uint8) * 255

    # line
    pos = np.random.uniform(0.72, 0.78)
    start = (0, int(pos*img.shape[0]))
    end = (img.shape[1], int(pos*img.shape[0]))

    # draw line
    # mean_rgb = int(0.3*img.mean())
    mean_rgb = int(np.random.uniform(0.25, 0.45)*img.mean())
    line_mask = cv2.line(line_mask,start,end,(mean_rgb,mean_rgb,mean_rgb),2)

    # blur line
    line_mask = cv2.GaussianBlur(line_mask, (7, 7), 0)

    # merge line to raw image
    img = np.minimum(img, line_mask)

    return img

def add_dotted_line(img):

    h, w, c = img.shape
    line_mask = np.ones((h,w,c), dtype=np.uint8) * 255

    # line
    pos = np.random.uniform(0.72, 0.78)
    start = (0, int(pos*img.shape[0]))
    end = (img.shape[1], int(pos*img.shape[0]))

    # draw line
    mean_rgb = np.random.randint(0, 10)
    line_mask = cv2.line(line_mask,start,end,(mean_rgb,mean_rgb,mean_rgb), 2)
    delete_idxs = []
    blank = np.random.randint(2, 7)
    for j in range(blank):
        delete_idxs += [i for i in range(j, w, 4+blank-2)]

    line_mask[start[1]-1, delete_idxs, :] = [[255,255,255]]*len(delete_idxs)
    line_mask[start[1], delete_idxs, :] = [[255,255,255]]*len(delete_idxs)
    line_mask[start[1]+1, delete_idxs, :] = [[255,255,255]]*len(delete_idxs)

    # blur line
    line_mask = cv2.GaussianBlur(line_mask, (5, 5), 0)

    # merge line to raw image
    img = np.minimum(img, line_mask)

    return img

def add_grid(img):

    h, w, c = img.shape

    mean_rgb = int(np.random.uniform(0.25, 0.45)*img.mean())

    # draw row
    pos1 = np.random.uniform(0, 0.06)
    rstart = [(0, int((i+pos1)*img.shape[0])) for i in [0, 0.25, 0.5, 0.75]]
    rend = [(img.shape[1], int((i+pos1)*img.shape[0])) for i in [0, 0.25, 0.5, 0.75]]
    row_mask = np.ones((h,w,c), dtype=np.uint8) * 255
    for i in range(4):
        row_mask = cv2.line(row_mask,rstart[i],rend[i],(mean_rgb,mean_rgb,mean_rgb),2)
    row_mask = cv2.GaussianBlur(row_mask, (7, 7), 0)

    #draw column
    pos2 = np.random.randint(0,5)
    cstart = [(i, 0) for i in range(pos2, w, 16)]
    cend = [(i, h) for i in range(pos2, w, 16)]
    column_mask = np.ones((h,w,c), dtype=np.uint8) * 255
    for i in range(len(cstart)):
        column_mask = cv2.line(column_mask,cstart[i],cend[i],(mean_rgb,mean_rgb,mean_rgb),2)
    column_mask = cv2.GaussianBlur(column_mask, (7, 7), 0)

    grid_mask = np.minimum(row_mask, column_mask)

    # merge line to raw image
    img = np.minimum(img, grid_mask)

    return img

def get_path(filename):
    s_char = filename[0]
    if s_char == 'g':
        return "images_gen/" + filename
    if s_char == 'w':
        return "grey_word/" + filename
    return "grey/" + filename

def augment():
    print("Start ...")
    anot = pd.read_csv("adjust_anot.csv").sample(frac=1, random_state=42).reset_index(drop=True)
    # anot.to_csv("augmented_anot.csv", index=False)
    anot = anot.to_dict('records')
    l_anot = anot[:1000000]
    d_anot = anot[1000000:2000000]
    g_anot = anot[2000000:3000000]
    _anot = anot[3000000:]
    print("1--------------------")
    for i in tqdm(range(len(l_anot))):
        fn = l_anot[i]["filename"]
        img = cv2.imread(get_path(fn))
        cv2.imwrite("augmented_images/"+fn, add_line(img))
    
    print("2--------------------")
    for i in tqdm(range(len(d_anot))):
        fn = d_anot[i]["filename"]
        img = cv2.imread(get_path(fn))
        cv2.imwrite("augmented_images/"+fn, add_dotted_line(img))
    
    print("3--------------------")
    for i in tqdm(range(len(g_anot))):
        fn = g_anot[i]["filename"]
        img = cv2.imread(get_path(fn))
        cv2.imwrite("augmented_images/"+fn, add_grid(img))
    
    print("4--------------------")
    for i in tqdm(range(len(_anot))):
        fn = _anot[i]["filename"]
        img = cv2.imread(get_path(fn))
        cv2.imwrite("augmented_images/"+fn, img)
    
    anot.sample(frac=1, random_state=0).reset_index(drop=True).to_csv("augmented_anot.csv", index=False)

if __name__ == "__main__":
    augment()
        



from dataclasses import dataclass
from vietocr.tool.predictor import Predictor
from vietocr.tool.config import Cfg
import torch

device = torch.device("cpu")

@dataclass
class Config:
    n_decoder_blocks: int = 3
    n_encoder_blocks: int = 3
    im_height: int = 32
    im_width: int = 768
    n_embd: int = 256
    feat_depth: int = 256
    n_heads: int = 8
    dropout: float = 0.4
    att_dropout: float = 0.0
    bias: bool = False
    pe_dropout: float = 0.2
    pe_maxlen: int = 100000
    max_seq_len: int = 128
    in_chanel: int = 3
    vocab_size: int = 233
    rank: int = 16

config = Config()

cfg = Cfg.load_config_from_name('vgg_transformer')
cfg['cnn']['pretrained']=False
cfg['device'] = device
predictor = Predictor(cfg).model


# app.py
import streamlit as st
from PIL import Image, ImageOps
import numpy as np
from transformers import AutoTokenizer, TrOCRProcessor
import torch
from VGGTransformer.jit import OCRModel
from VisionEncoderDecoder.jit import TrOCR
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, Normalize

device = "cuda:2"

def process_input(image):
    img = image.convert('RGB')
    image_height = 32

    w, h = img.size
    new_w = int(image_height*w/h)

    img = img.resize((new_w, image_height), Image.LANCZOS)
    img = ImageOps.expand(img, border=(0, 0, 768 - new_w, 0), fill='white')
    img = np.asarray(img).transpose(2,0, 1)
    img = img/255
    return img, new_w

class Tokenizer():
    def __init__(self):
        vocab = ['a', 'A', 'à', 'À', 'ả', 'Ả', 'ã', 'Ã', 'á', 'Á', 'ạ', 'Ạ', 'ă', 'Ă', 'ằ', 'Ằ', 'ẳ', 'Ẳ', 'ẵ', 'Ẵ', 'ắ', 'Ắ', 'ặ', 'Ặ', 'â', 'Â', 'ầ', 'Ầ', 'ẩ', 'Ẩ', 'ẫ', 'Ẫ', 'ấ', 'Ấ', 'ậ', 'Ậ', 'b', 'B', 'c', 'C', 'd', 'D', 'đ', 'Đ', 'e', 'E', 'è', 'È', 'ẻ', 'Ẻ', 'ẽ', 'Ẽ', 'é', 'É', 'ẹ', 'Ẹ', 'ê', 'Ê', 'ề', 'Ề', 'ể', 'Ể', 'ễ', 'Ễ', 'ế', 'Ế', 'ệ', 'Ệ', 'f', 'F', 'g', 'G', 'h', 'H', 'i', 'I', 'ì', 'Ì', 'ỉ', 'Ỉ', 'ĩ', 'Ĩ', 'í', 'Í', 'ị', 'Ị', 'j', 'J', 'k', 'K', 'l', 'L', 'm', 'M', 'n', 'N', 'o', 'O', 'ò', 'Ò', 'ỏ', 'Ỏ', 'õ', 'Õ', 'ó', 'Ó', 'ọ', 'Ọ', 'ô', 'Ô', 'ồ', 'Ồ', 'ổ', 'Ổ', 'ỗ', 'Ỗ', 'ố', 'Ố', 'ộ', 'Ộ', 'ơ', 'Ơ', 'ờ', 'Ờ', 'ở', 'Ở', 'ỡ', 'Ỡ', 'ớ', 'Ớ', 'ợ', 'Ợ', 'p', 'P', 'q', 'Q', 'r', 'R', 's', 'S', 't', 'T', 'u', 'U', 'ù', 'Ù', 'ủ', 'Ủ', 'ũ', 'Ũ', 'ú', 'Ú', 'ụ', 'Ụ', 'ư', 'Ư', 'ừ', 'Ừ', 'ử', 'Ử', 'ữ', 'Ữ', 'ứ', 'Ứ', 'ự', 'Ự', 'v', 'V', 'w', 'W', 'x', 'X', 'y', 'Y', 'ỳ', 'Ỳ', 'ỷ', 'Ỷ', 'ỹ', 'Ỹ', 'ý', 'Ý', 'ỵ', 'Ỵ', 'z', 'Z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`', '{', '|', '}', '~', ' ']
        vocab = ["pad", "<sos>", "<eos>", "<unk>"] + vocab

        self.c_vocab = dict(zip(vocab, np.arange(len(vocab))))
        self.reverse_vocab = list(self.c_vocab.items())
        self.max_seq_len = 128

    def tokenize(self, seq):
        res = []
        for c in ['<sos>'] + seq + ['<eos>']:
            try:
                res.append(self.c_vocab[c])
            except:
                res.append(self.c_vocab["<unk>"])
        mask = [1]*len(res)
        if len(res) < self.max_seq_len:
            n_pad = self.max_seq_len-len(res)+1
            res += [self.c_vocab["pad"]]*n_pad
            mask += [0]*n_pad
        elif len(res) > self.max_seq_len:
            s_pad = self.max_seq_len
            res = res[:s_pad] + [self.c_vocab["<eos>"]]
            mask = mask[:s_pad+1]
        else:
            res += [self.c_vocab["pad"]]
            mask += [0]
        return np.array(res), np.logical_not(mask)

    def reverse_tokens(self, tokens):
        res = []
        for token in tokens:
            if token in [self.c_vocab["<unk>"], self.c_vocab["pad"], self.c_vocab["<sos>"]]:
                continue
            if token == self.c_vocab["<eos>"]:
                break
            res.append(self.reverse_vocab[token][0])
            
        return ''.join(res)

    def reverse_tokens_special(self, tokens):
        res = []
        for token in tokens:
            res.append(self.reverse_vocab[token][0])
            # if token == self.c_vocab["<eos>"]:
            #     break
            
        return res

if 'session_state' not in st.session_state:
    st.session_state.session_state = {
        'v_result': None,
        'va_result': None,
        'tr_result': None,
        'tra_result': None,
    }

def perform_inference(img, model, label):
    pred, token, prob, conflict = None, None, None, None
    
    if model is not None:
        pred, token, prob, conflict = model.predict_topk(img)
        st.session_state[f'{label.lower().replace(" ", "_")}_result'] = {
            'pred': pred,
            'token': token,
            'prob': prob,
            'conflict': conflict,
        }
        if label == "VGG Transformer":
            st.write("VGG Transformer:\t\t\t" + tkz.reverse_tokens(pred[0]))
        elif label == "VGG Transformer with Rethinking Head":
            st.write("VGG Transformer with Rethinking Head:\t" + tkz.reverse_tokens(pred[0]))
        elif label == "TrOCR":
            st.write("TrOCR:\t\t\t\t\t" + tokenizer.batch_decode(pred, skip_special_tokens=True)[0])
        else:st.write("TrOCR with Rethinking Head:\t\t" + tokenizer.batch_decode(pred, skip_special_tokens=True)[0])


    return {
        'pred': pred,
        'token': token,
        'prob': prob,
        'conflict': conflict,
    }

def inference(img, opt):
    img1, W = process_input(img)
    img1 = torch.FloatTensor(img1[:,:W,:]).unsqueeze(0).to(device)
    img2 = processor(img, return_tensors="pt").pixel_values.to(device)
    
    if opt == 0:
        st.session_state.v_result = perform_inference(img1, vgg_base, "VGG Transformer")
    elif opt == 1:
        st.session_state.va_result = perform_inference(img1, vgg_adapter, "VGG Transformer with Rethinking Head")
    elif opt == 2:
        st.session_state.tr_result = perform_inference(img2, trocr_base, "TrOCR")
    else:
        st.session_state.tra_result = perform_inference(img2, trocr_base, "TrOCR with Rethinking Head")


def plot_entire(token, prob, conflict, topk=5):
    if conflict is not None:
        if conflict.shape[-1] < 500:
            weights = prob[:topk, :]
            token = token[:topk, :]

            fig, ax = plt.subplots(figsize=(20, 2.5))

            # Create a custom colormap that transitions from white to 'viridis'
            viridis_white = plt.cm.Blues(np.linspace(0, 1, 256))
            viridis_white[:, :3] = 1 - (1 - viridis_white[:, :3]) * 0.7  # Adjust the brightness

            # Create a LinearSegmentedColormap using the custom colormap
            cmap = LinearSegmentedColormap.from_list('custom_cmap', viridis_white)

            # Set the normalization to map the weights to the range [0, 1]
            norm = Normalize(vmin=0, vmax=1)

            # Use the created colormap in imshow with the specified norm
            cax = ax.imshow(weights, cmap=cmap, interpolation='none', norm=norm)

            for i in range(token.shape[0]):
                text = tkz.reverse_tokens_special(token[i])
                for j in range(len(text)):
                    char = text[j]
                    if char == "<unk>": char = "/u"
                    if char == "<eos>": char = "/e"
                    ax.text(j, i, char, ha='center', va='center', fontsize=14)

            # Add colorbar to display the color scale
            cbar = plt.colorbar(cax)

            st.pyplot(fig)

            pos = st.number_input("Select position", value=0, step=1)
            topk_ = st.slider('Select topk_:', min_value=1, max_value=100, value=10, step=1)
            plot_pos(conflict, pos, topk_)
        
        else:
            weights = prob.T[:topk, :]

            fig, ax = plt.subplots(figsize=(17, 5))

            # Create a custom colormap that transitions from white to 'viridis'
            viridis_white = plt.cm.Blues(np.linspace(0, 1, 256))
            viridis_white[:, :3] = 1 - (1 - viridis_white[:, :3]) * 0.7  # Adjust the brightness

            # Create a LinearSegmentedColormap using the custom colormap
            cmap = LinearSegmentedColormap.from_list('custom_cmap', viridis_white)

            # Set the normalization to map the weights to the range [0, 1]
            norm = Normalize(vmin=0, vmax=1)

            # Use the created colormap in imshow with the specified norm
            cax = ax.imshow(weights, cmap=cmap, interpolation='none', norm=norm)

            for i in range(token.shape[1]):
                text = tokenizer.batch_decode(token[:, i:i+1])
                for j in range(len(text)):
                    char = text[j]
                    ax.text(j, i, char, ha='center', va='center', fontsize=14)

            # Add colorbar to display the color scale
            cbar = plt.colorbar(cax)
            st.pyplot(fig)
            pos = st.number_input("Select position", value=0, step=1)
            topk_ = st.slider('Select topk_:', min_value=1, max_value=100, value=10, step=1)
            plot_pos(conflict, pos, topk_)
    else:
        print("here")

def plot_pos(conflict, pos, topk_):
    if conflict is not None:
        if conflict.shape[-1] < 500:
            categories = tkz.reverse_tokens_special(np.arange(233))
            categories = ['/p', '/s/', '/s', '/u'] + categories[4:]
            values = conflict[pos][0]

            sorted_indices = sorted(range(len(values)), key=lambda k: values[k], reverse=True)
            categories = [categories[i] for i in sorted_indices]
            values = [values[i] for i in sorted_indices]

            # Vẽ biểu đồ cột
            fig = plt.figure(figsize=(8, 4))
            bars = plt.bar(categories[:topk_], values[:topk_])

            # Đặt tên cho trục x và y
            plt.xlabel('Labels')
            plt.ylabel('Probs')

            plt.ylim(0, 1)
            # Hiển thị biểu đồ
            st.pyplot(fig)

@st.cache_resource()
def load_model(mtype, device):
    if mtype == 0:
        return OCRModel(False, device)
    if mtype == 1:
        return OCRModel(True, device)
    if mtype == 2:
        return TrOCR(False, device)
    if mtype == 3:
        return TrOCR(True, device)

@st.cache_resource()
def get_tkz(ttype):
    if ttype == 0:
        return AutoTokenizer.from_pretrained("vinai/bartpho-syllable-base")
    if ttype == 1:
        return TrOCRProcessor.from_pretrained("microsoft/trocr-small-handwritten")
    if ttype == 2:
        return Tokenizer()

vgg_base = load_model(0, device)
vgg_adapter = load_model(1, device)
trocr_base = load_model(2, device)
trcocr_adapter = load_model(3, device)

tokenizer = get_tkz(0)
processor = get_tkz(1)
tkz = get_tkz(2)

def main():
    st.title("Vietnamese Handwriting Text Recognition")
    img = None
    images = st.file_uploader("Upload image", type=["jpg", "jpeg", "png"], accept_multiple_files=True)
    if images is not None:
        for image in images:
            img = Image.open(image).convert("RGB")
            st.image(img, use_column_width=True)

    col1, col2, col3, col4, col5 = st.columns(5)
    with col1:
        option_a = st.checkbox("Infer by Vgg Transformer")
    with col2:
        option_b = st.checkbox("Infer by Vgg Transformer with Rethinking Head")
    with col3:
        option_c = st.checkbox("Infer by TrOCR")
    with col4:
        option_d = st.checkbox("Infer by TrOCR with Rethinking Head")
    with col5:
        select_all = st.button("Select all")

    # Nếu người dùng nhấn nút "Chọn tất cả", đặt giá trị của các ô kiểm thành True
    if select_all:
        option_a = option_b = option_c = option_d = True

    # Nút "Infer" để thực hiện công việc dựa trên các tùy chọn đã chọn
    if st.button("Infer"):
        if option_a:    inference(img, 0)
        if option_b:    inference(img, 1)
        if option_c:    inference(img, 2)
        if option_d:    inference(img, 3)

    st.markdown("## Visualize probabilities")
    selected_model = st.radio('Select model:', ('VGG Transformer', 'VGG Transformer with Rethinking Head', 'TrOCR', 'TrOCR with Rethinking Head'))
    topk = st.slider('Select topk:', min_value=1, max_value=50, value=5, step=1)
    result_key = selected_model.lower().replace(" ", "_") + '_result'
    result = st.session_state.get(result_key, {})

    if result:
        plot_entire(result['token'], result['prob'], result['conflict'], topk)


    if st.button("Clear"):
        pass


if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
from torch.nn import functional as F

def infer(pixel_values, net, max_seq_length=35, sos_token=0, eos_token=2, adapter=False):
    "data: BxCXHxW"
    net.eval()
    device = pixel_values.device
    net.to(device)
    if not adapter:
        model = net
    else:
        model = net.basemodel

    with torch.no_grad():
        encoder_output = model.forward_encode(pixel_values)
    
        start_ids = torch.LongTensor([[sos_token]]*pixel_values.shape[0]).to(device)

        max_length = 0

        while max_length <= max_seq_length and not all(start_ids[:, -1] == eos_token):
            output = model.forward_decode(start_ids, encoder_output)
            if adapter:
                output = net.forward_adapt(output)
            output = F.softmax(output[:,-1,:], dim=-1)

            start_ids = torch.cat([start_ids, output.argmax(dim=-1).unsqueeze(1)], dim=-1)

            max_length += 1
    
    return start_ids

def param_to_update(net):
        encoder_param = []
        decoder_param = []
        for name, param in net.named_parameters():
            param = param.float()
            param.requires_grad = True
            if "backbone" in name:
                encoder_param.append(param)
            else:
                decoder_param.append(param)
        return encoder_param, decoder_param


from model import *
from utils import *
from dataset import *
from torch.utils.data import DataLoader
import pandas as pd
from tqdm import tqdm
from evaluate import load
import wandb
from config import *
from torch.utils.data.distributed import DistributedSampler
import torch.distributed as dist
import argparse
import os


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--local_rank", 
                        type=int, 
                        default=os.getenv('LOCAL_RANK', -1), 
                        help="Local rank. Necessary for using the torch.distributed.launch utility.")
    args = parser.parse_args()
    
    local_rank = args.local_rank
    torch.distributed.init_process_group(backend="nccl", rank = local_rank)
    torch.cuda.set_device(local_rank)
    device = torch.device(local_rank)

    #define model
    torch.set_default_dtype(torch.float32)
    torch.autograd.set_detect_anomaly(True)
    model = AdapterVNTrOCR(config)
    # model = nn.DataParallel(model)
    model.to(device)
    model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[local_rank], output_device=local_rank, find_unused_parameters=True)

    #data loader
    anot = pd.read_csv("concated_anot.csv").sample(frac=1, random_state=0).reset_index(drop=True)
    model.module.basemodel.load_state_dict(torch.load("VisionEncoderDecoder/weights/cp_vision_encoder_decoder_augmented_data.pt", map_location=device))

    traindataset = CustomDataset(anot[:-10000], config)
    train_sampler = DistributedSampler(dataset=traindataset)
    trainloader = DataLoader(dataset=traindataset, batch_size=16, generator=torch.Generator(device="cpu"), num_workers=2, sampler = train_sampler, pin_memory = True)
    valdataset = CustomDataset(anot[-10000:], config)
    valloader = DataLoader(dataset=valdataset, batch_size=96, generator=torch.Generator(device="cpu"), num_workers=2, pin_memory = True)

    #optimizer
    param1, param2 = param_to_update(model)
    optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.0001},
                                {'params': param2, 'lr': 0.0001}], weight_decay=1e-5)

    #loss function
    # lsLoss = LabelSmoothingLoss(233, padding_idx=0, smoothing=0.1)

    #utils
    cer = load("cer")

    #number of epochs
    n_epoch = 15
    #checkpoint
    best_cer = 1

    # wandb init
    # wandb.login(key = '')
    # wandb.login(key = '')
    # wandb.init(
    #   project="VN_HTR",
    #   name=f"adapter_vision_encoder_decoder_add",
    #   config={
    #       "learning_rate": 0.0001,
    #       "architecture": "v_t",
    #       "dataset": "gen",
    #       "epochs": 10,
    #   }
    # )

    print(local_rank)
    print("Start .....")
    dist.barrier()
    torch.backends.cudnn.benchmark = True
    for epoch in range(n_epoch):
        dist.barrier()        
        # map_location = {'cuda:%d' % 0: 'cuda:%d' % local_rank}
        if epoch != 0:
            state_dict = torch.load('/mnt/disk4/VN_HTR/VN_HTR/VisionEncoderDecoder/new_check_point/tmp.pt', map_location = 'cuda:{}'.format(local_rank))
            model.module.load_state_dict(state_dict)
        
        print(f"epoch {epoch} ----------------------------")
        if epoch < 1:
            optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00005},
                            {'params': param2, 'lr': 0.00005}], weight_decay=1e-5)
        # elif epoch < 3:
        #     optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00001},
        #                     {'params': param2, 'lr': 0.00001}], weight_decay=1e-5)
        else:
            optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00001},
                            {'params': param2, 'lr': 0.00001}], weight_decay=1e-5)
        model.train()
        train_loss = 0
        torch.cuda.empty_cache()
        for batch in tqdm(trainloader):
            pixel_values = batch["pixel_values"].to(device)
            input_ids = batch["input_ids"].to(device)
            att_mask = batch["att_mask"].to(device)

            optimizer.zero_grad()

            # forward
            loss = model(pixel_values, input_ids, att_mask)
            
            loss.backward()
            
            torch.nn.utils.clip_grad_norm_(model.parameters(), 4)
            
            optimizer.step()

            train_loss += loss.item()
        else:
            dist.barrier()
            print(f"train loss: {train_loss/len(trainloader)}")
            # torch.save(model.state_dict(), "source/weights/last_vgg_transformer.pt")
            if True:
                with torch.no_grad():
                    model.eval()
                    predictions, references = [], []
                    torch.cuda.empty_cache()
                    for batch in tqdm(valloader):
                        pixel_values = batch["pixel_values"].to(device)
                        input_ids = batch["input_ids"].to(device)

                        preds = infer(pixel_values, model.module, adapter=True)

                        predictions += valdataset.tokenizer.batch_decode(preds, skip_special_tokens=True)
                        references += valdataset.tokenizer.batch_decode(input_ids, skip_special_tokens=True)
                    for i in range(len(references)):
                        if references[i] == "":
                            references[i] = " "
                    cer_score = 1
                    try:
                        cer_score = cer.compute(predictions=predictions, references=references, concatenate_texts=True)
                    except:
                        print("error when caculate cer score")
                    # wandb.log({"train_loss": train_loss/len(trainloader), "cer": cer_score})
                    if local_rank == 0:
                        torch.save(model.module.state_dict(), "/mnt/disk4/VN_HTR/VN_HTR/VisionEncoderDecoder/new_check_point/tmp.pt")

                    if cer_score < best_cer and local_rank == 0:
                        best_cer = cer_score
                        torch.save(model.module.state_dict(), "VisionEncoderDecoder/weights/cp_adapter_add.pt")

                    print(f'Character Error Rate: {cer_score}')
    
    # wandb.finish()


import math
import torch.nn as nn
import torch
from torch.nn import functional as F
from torchvision import models
import math
from config import config, predictor, device
from torch.nn import Module, ModuleList
from torch.nn import Conv2d, InstanceNorm2d, Dropout, Dropout2d
from torch.nn import ReLU
from torch.nn.functional import pad
import random

class Rethinking(nn.Module):
    def __init__(self, config, masked=True):
        super().__init__()
        self.n_embd = config.vocab_size
        self.block_size = config.max_seq_len
        self.c_attn = nn.Linear(self.n_embd, config.rank*3, bias=config.bias)
        self.c_proj = nn.Linear(config.rank, self.n_embd, bias=config.bias)
        self.attn_dropout = nn.Dropout(0)
        self.resid_dropout = nn.Dropout(0.1)
        self.gelu = nn.GELU()
        self.masked = masked


    # def forward(self, x, mask):
    def forward(self, x):
        B, T, C = x.shape                                               #batch_size, block_size, n_embd
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)

        # q = q.view(B, T, self.n_heads, C//self.n_heads).transpose(1,2)  # (B, n_heads, T, h_size)
        # k = k.view(B, T, self.n_heads, C//self.n_heads).transpose(1,2)
        # v = v.view(B, T, self.n_heads, C//self.n_heads).transpose(1,2)

        att = q@k.transpose(-2,-1)*(1/math.sqrt(k.size(-1)))            # (B, T, T)
        if self.masked:
            mask = torch.tril(torch.ones((T, T))).view(1, T, T).to(x.device)
            att = att.masked_fill_(mask == 0, -float('inf'))
        # att = att + mask

        att = F.softmax(att, dim=-1)
        att = self.attn_dropout(att)
        y = att@v                                                       #(B, T, n_embd)
        # y = y.transpose(1,2).contiguous().view(B,T,C)
        y = self.resid_dropout(self.c_proj(y))
        return self.gelu(y)

class TransformerLM(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.d_model = config.n_embd
        
        self.embed_tgt = predictor.transformer.embed_tgt
        self.pos_enc = predictor.transformer.pos_enc
        self.transformer = predictor.transformer.transformer
        self.fc = predictor.transformer.fc
        
    def forward(self, src, tgt, src_key_padding_mask=None, tgt_key_padding_mask=None, memory_key_padding_mask=None):
        """
        Shape:
            - src: (W, N, C)
            - tgt: (T, N) 
            - src_key_padding_mask: (N, S)
            - tgt_key_padding_mask: (N, T)
            - memory_key_padding_mask: (N, S)
            - output: (N, T, E)
            
        """
        tgt_mask = self.gen_nopeek_mask(tgt.shape[0]).to(src.device)
        
        src = self.pos_enc(src*math.sqrt(self.d_model))
#        src = self.learned_pos_enc(src*math.sqrt(self.d_model))

        tgt = self.pos_enc(self.embed_tgt(tgt) * math.sqrt(self.d_model))
        
        output = self.transformer(src, tgt, tgt_mask=tgt_mask, src_key_padding_mask=src_key_padding_mask,
                                  tgt_key_padding_mask=tgt_key_padding_mask, memory_key_padding_mask=memory_key_padding_mask)
#        output = rearrange(output, 't n e -> n t e')
        output = output.transpose(0, 1)
        return self.fc(output)
    
    def gen_nopeek_mask(self, length):
        mask = (torch.triu(torch.ones(length, length)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask
    
    def forward_encoder(self, src):
        src = self.pos_enc(src*math.sqrt(self.d_model))
        memory = self.transformer.encoder(src)
        return memory
    
    def forward_decoder(self, tgt, memory):
        tgt_mask = self.gen_nopeek_mask(tgt.shape[0]).to(tgt.device)
        tgt = self.pos_enc(self.embed_tgt(tgt) * math.sqrt(self.d_model))

        output = self.transformer.decoder(tgt, memory, tgt_mask=tgt_mask)
        output = output.transpose(0, 1)

        return self.fc(output), memory

class PositionalEncoding1D(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dropout = nn.Dropout(p=config.pe_dropout)

        position = torch.arange(config.pe_maxlen).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, config.n_embd, 2) * (-math.log(10000.0) / config.n_embd))
        self.pe = torch.zeros(config.pe_maxlen, 1, config.n_embd).to(device)
        self.pe[:, 0, 0::2] = torch.sin(position * div_term)
        self.pe[:, 0, 1::2] = torch.cos(position * div_term)

    def forward(self, x):
        """
        Arguments:
            x: Tensor, shape ``[batch_size, seq_len, embedding_dim]``
        """
        x = x + self.pe[:x.size(1)].transpose(0,1)
        return self.dropout(x)

class Resnet34(nn.Module):
    def __init__(self, config):
        super().__init__()
#         self.backbone = ConvEmbeddingGC()
        self.backbone = models.resnet34(pretrained=True)
        self.backbone = nn.Sequential(*list(self.backbone.children())[:-2])
        self.project = nn.Conv2d(512, config.n_embd, kernel_size=1)
#         self.pe2d = PositionalEncoding2D(config)
#         self.att = EncoderTrans(config)

    def forward(self, x):
        x = self.backbone(x)            #(B, feat_depth, H, H)
        x = self.project(x)
#         x = self.pe2d(x)
        x = torch.flatten(x, start_dim=-2).transpose(1,2).contiguous() #(B, H*H, n_emb)
        return x



class LabelSmoothingLoss(nn.Module):
    def __init__(self, classes, padding_idx, smoothing=0.0, dim=-1):
        super(LabelSmoothingLoss, self).__init__()
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing
        self.cls = classes
        self.dim = dim
        self.padding_idx = padding_idx

    def forward(self, pred, target):
        pred = pred.view(-1, pred.size(2))
        target = target.reshape(-1)
        pred = pred.log_softmax(dim=self.dim)
        with torch.no_grad():
            # true_dist = pred.data.clone()
            true_dist = torch.zeros_like(pred)
            true_dist.fill_(self.smoothing / (self.cls - 2))
            true_dist.scatter_(1, target.data.unsqueeze(1), self.confidence)
            true_dist[:, self.padding_idx] = 0
            mask = torch.nonzero(target.data == self.padding_idx, as_tuple=False)
            if mask.dim() > 0:
                true_dist.index_fill_(0, mask.squeeze(), 0.0)
            
        return torch.mean(torch.sum(-true_dist * pred, dim=self.dim))
    
class MultiAspectGCAttention(nn.Module):

    def __init__(self,
                 inplanes,
                 ratio=0.065,
                 headers=4,
                 pooling_type='att',
                 att_scale=False,
                 fusion_type='channel_concat'):
        super().__init__()
        assert pooling_type in ['avg', 'att']

        assert fusion_type in ['channel_add', 'channel_mul', 'channel_concat']
        assert inplanes % headers == 0 and inplanes >= 8  # inplanes must be divided by headers evenly

        self.headers = headers
        self.inplanes = inplanes
        self.ratio = ratio
        self.planes = int(inplanes * ratio)
        self.pooling_type = pooling_type
        self.fusion_type = fusion_type
        self.att_scale = False

        self.single_header_inplanes = int(inplanes / headers)

        if pooling_type == 'att':
            self.conv_mask = nn.Conv2d(self.single_header_inplanes, 1, kernel_size=1)
            self.softmax = nn.Softmax(dim=2)
        else:
            self.avg_pool = nn.AdaptiveAvgPool2d(1)

        if fusion_type == 'channel_add':
            self.channel_add_conv = nn.Sequential(
                nn.Conv2d(self.inplanes, self.planes, kernel_size=1),
                nn.LayerNorm([self.planes, 1, 1]),
                nn.ReLU(inplace=True),
                nn.Conv2d(self.planes, self.inplanes, kernel_size=1))
        elif fusion_type == 'channel_concat':
            self.channel_concat_conv = nn.Sequential(
                nn.Conv2d(self.inplanes, self.planes, kernel_size=1),
                nn.LayerNorm([self.planes, 1, 1]),
                nn.ReLU(inplace=True),
                nn.Conv2d(self.planes, self.inplanes, kernel_size=1))
            # for concat
            self.cat_conv = nn.Conv2d(2 * self.inplanes, self.inplanes, kernel_size=1)
        elif fusion_type == 'channel_mul':
            self.channel_mul_conv = nn.Sequential(
                nn.Conv2d(self.inplanes, self.planes, kernel_size=1),
                nn.LayerNorm([self.planes, 1, 1]),
                nn.ReLU(inplace=True),
                nn.Conv2d(self.planes, self.inplanes, kernel_size=1))

    def spatial_pool(self, x):
        batch, channel, height, width = x.size()
        if self.pooling_type == 'att':
            # [N*headers, C', H , W] C = headers * C'
            x = x.view(batch * self.headers, self.single_header_inplanes, height, width)
            input_x = x

            # [N*headers, C', H * W] C = headers * C'
            # input_x = input_x.view(batch, channel, height * width)
            input_x = input_x.view(batch * self.headers, self.single_header_inplanes, height * width)

            # [N*headers, 1, C', H * W]
            input_x = input_x.unsqueeze(1)
            # [N*headers, 1, H, W]
            context_mask = self.conv_mask(x)
            # [N*headers, 1, H * W]
            context_mask = context_mask.view(batch * self.headers, 1, height * width)

            # scale variance
            if self.att_scale and self.headers > 1:
                context_mask = context_mask / torch.sqrt(torch.tensor(self.single_header_inplanes))

            # [N*headers, 1, H * W]
            context_mask = self.softmax(context_mask)

            # [N*headers, 1, H * W, 1]
            context_mask = context_mask.unsqueeze(-1)
            # [N*headers, 1, C', 1] = [N*headers, 1, C', H * W] * [N*headers, 1, H * W, 1]
            context = torch.matmul(input_x, context_mask)

            # [N, headers * C', 1, 1]
            context = context.view(batch, self.headers * self.single_header_inplanes, 1, 1)
        else:
            # [N, C, 1, 1]
            context = self.avg_pool(x)

        return context

    def forward(self, x):
        # [N, C, 1, 1]
        context = self.spatial_pool(x)

        out = x

        if self.fusion_type == 'channel_mul':
            # [N, C, 1, 1]
            channel_mul_term = torch.sigmoid(self.channel_mul_conv(context))
            out = out * channel_mul_term
        elif self.fusion_type == 'channel_add':
            # [N, C, 1, 1]
            channel_add_term = self.channel_add_conv(context)
            out = out + channel_add_term
        else:
            # [N, C, 1, 1]
            channel_concat_term = self.channel_concat_conv(context)

            # use concat
            _, C1, _, _ = channel_concat_term.shape
            N, C2, H, W = out.shape

            out = torch.cat([out, channel_concat_term.expand(-1, -1, H, W)], dim=1)
            out = self.cat_conv(out)
            out = nn.functional.layer_norm(out, [self.inplanes, H, W])
            out = nn.functional.relu(out)

        return out

def conv3x3(in_planes, out_planes, stride=1):
    # "3x3 convolution with padding"
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, use_gcb=False, gcb_config=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes, momentum=0.9)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes, momentum=0.9)
        self.downsample = downsample
        self.stride = stride
        self.use_gcb = use_gcb

        if self.use_gcb:
            gcb_ratio = gcb_config['ratio']
            gcb_headers = gcb_config['headers']
            att_scale = gcb_config['att_scale']
            fusion_type = gcb_config['fusion_type']
            self.context_block = MultiAspectGCAttention(inplanes=planes,
                                                        ratio=gcb_ratio,
                                                        headers=gcb_headers,
                                                        att_scale=att_scale,
                                                        fusion_type=fusion_type)

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.use_gcb:
            out = self.context_block(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out = out + residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = conv1x1(inplanes, planes)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = conv3x3(planes, planes, stride)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = conv1x1(planes, planes * self.expansion)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out = out + identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, layers, zero_init_residual=False, gcb=None, in_channels=1):
        super(ResNet, self).__init__()
        gcb_config = gcb

        self.inplanes = 64
        self.conv1 = nn.Conv2d(in_channels, 32, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(32)
        self.relu1 = nn.ReLU(inplace=True)

        self.conv2 = nn.Conv2d(32, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(64)
        self.relu2 = nn.ReLU(inplace=True)

        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)

        self.layer1 = self._make_layer(block, 128, layers[0], stride=1, gcb_config=gcb_config,
                                       use_gcb=gcb_config['layers'][0])

        self.conv3 = nn.Conv2d(128, 128, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(128)
        self.relu3 = nn.ReLU(inplace=True)

        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)

        self.layer2 = self._make_layer(block, 128, layers[1], stride=1, gcb_config=gcb_config,
                                       use_gcb=gcb_config['layers'][1])

        self.conv4 = nn.Conv2d(128, 128, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn4 = nn.BatchNorm2d(128)
        self.relu4 = nn.ReLU(inplace=True)

        self.maxpool3 = nn.MaxPool2d(kernel_size=(2, 2), stride=(2, 2))

        self.layer3 = self._make_layer(block, 256, layers[2], stride=1, gcb_config=gcb_config,
                                       use_gcb=gcb_config['layers'][2])

        self.conv5 = nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn5 = nn.BatchNorm2d(256)
        self.relu5 = nn.ReLU(inplace=True)
        
        # self.maxpool4 = nn.MaxPool2d(kernel_size=(2, 2), stride=(2, 2))

        self.layer4 = self._make_layer(block, 256, layers[3], stride=1, gcb_config=gcb_config,
                                       use_gcb=gcb_config['layers'][3])

        self.conv6 = nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn6 = nn.BatchNorm2d(256)
        self.relu6 = nn.ReLU(inplace=True)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, use_gcb=False, gcb_config=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, use_gcb=use_gcb, gcb_config=gcb_config))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu1(x)

        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu2(x)
        x = self.maxpool1(x)
        x = self.layer1(x)

        x = self.conv3(x)
        x = self.bn3(x)
        x = self.relu3(x)
        x = self.maxpool2(x)
        x = self.layer2(x)

        x = self.conv4(x)
        x = self.bn4(x)
        x = self.relu4(x)
        x = self.maxpool3(x)
        x = self.layer3(x)

        x = self.conv5(x)
        x = self.bn5(x)
        x = self.relu5(x)
        # x = self.maxpool4(x)
        x = self.layer4(x)

        x = self.conv6(x)
        x = self.bn6(x)
        x = self.relu6(x)
        return x


def resnet50(gcb_kwargs, in_channels=1):
    model = ResNet(BasicBlock, [1, 2, 5, 3], gcb=gcb_kwargs, in_channels=in_channels)
    return model


class ConvEmbeddingGC(nn.Module):

    def __init__(self, in_channels=3):
        super().__init__()
        gcb_kwargs = {
                    "ratio": 0.1,
                    "headers": 4,
                    "att_scale": True,
                    "fusion_type": "channel_concat",
                    "layers":[False, True, True, True]
                }
        self.cnn = resnet50(gcb_kwargs, in_channels=in_channels)

    def forward(self, x):
        feature = self.cnn(x)
        b, c, h, w = feature.shape  # （B， C， H/8, W/8)
        feature = feature.view(b, c, h * w)
        feature = feature.permute((0, 2, 1))
        return feature

class DepthSepConv2D(Module):
    def __init__(self, in_channels, out_channels, kernel_size, activation=None, padding=True, stride=(1, 1), dilation=(1, 1)):
        super(DepthSepConv2D, self).__init__()

        self.padding = None

        if padding:
            if padding is True:
                padding = [int((k - 1) / 2) for k in kernel_size]
                if kernel_size[0] % 2 == 0 or kernel_size[1] % 2 == 0:
                    padding_h = kernel_size[1] - 1
                    padding_w = kernel_size[0] - 1
                    self.padding = [padding_h//2, padding_h-padding_h//2, padding_w//2, padding_w-padding_w//2]
                    padding = (0, 0)

        else:
            padding = (0, 0)
        self.depth_conv = Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=kernel_size, dilation=dilation, stride=stride, padding=padding, groups=in_channels)
        self.point_conv = Conv2d(in_channels=in_channels, out_channels=out_channels, dilation=dilation, kernel_size=(1, 1))
        self.activation = activation

    def forward(self, x):
        x = self.depth_conv(x)
        if self.padding:
            x = pad(x, self.padding)
        if self.activation:
            x = self.activation(x)
        x = self.point_conv(x)
        return x


class MixDropout(Module):
    def __init__(self, dropout_proba=0.4, dropout2d_proba=0.2):
        super(MixDropout, self).__init__()

        self.dropout = Dropout(dropout_proba)
        self.dropout2d = Dropout2d(dropout2d_proba)

    def forward(self, x):
        if random.random() < 0.5:
            return self.dropout(x)
        return self.dropout2d(x)


class FCN_Encoder(Module):
    def __init__(self, params):
        super(FCN_Encoder, self).__init__()

        self.dropout = params["dropout"]

        self.init_blocks = ModuleList([
            ConvBlock(params["input_channels"], 16, stride=(1, 1), dropout=self.dropout),
            ConvBlock(16, 32, stride=(2, 2), dropout=self.dropout),
            ConvBlock(32, 64, stride=(2, 2), dropout=self.dropout),
            ConvBlock(64, 128, stride=(2, 2), dropout=self.dropout),
            ConvBlock(128, 128, stride=(2, 1), dropout=self.dropout),
            ConvBlock(128, 128, stride=(2, 1), dropout=self.dropout),
        ])
        self.blocks = ModuleList([
            DSCBlock(128, 128, pool=(1, 1), dropout=self.dropout),
            DSCBlock(128, 128, pool=(1, 1), dropout=self.dropout),
            DSCBlock(128, 128, pool=(1, 1), dropout=self.dropout),
            DSCBlock(128, 256, pool=(1, 1), dropout=self.dropout),
        ])

    def forward(self, x):
        for b in self.init_blocks:
            x = b(x)
        for b in self.blocks:
            xt = b(x)
            x = x + xt if x.size() == xt.size() else xt
        bz, c, h, w = x.shape  # （B， C， H/8, W/8)
        x = x.view(bz, c, h * w)
        x = x.permute((0, 2, 1))

        return x


class ConvBlock(Module):

    def __init__(self, in_, out_, stride=(1, 1), k=3, activation=ReLU, dropout=0.4):
        super(ConvBlock, self).__init__()

        self.activation = activation()
        self.conv1 = Conv2d(in_channels=in_, out_channels=out_, kernel_size=k, padding=k // 2)
        self.conv2 = Conv2d(in_channels=out_, out_channels=out_, kernel_size=k, padding=k // 2)
        self.conv3 = Conv2d(out_, out_, kernel_size=(3, 3), padding=(1, 1), stride=stride)
        self.norm_layer = InstanceNorm2d(out_, eps=0.001, momentum=0.99, track_running_stats=False)
        self.dropout = MixDropout(dropout_proba=dropout, dropout2d_proba=dropout / 2)

    def forward(self, x):
        pos = random.randint(1, 3)
        x = self.conv1(x)
        x = self.activation(x)

        if pos == 1:
            x = self.dropout(x)

        x = self.conv2(x)
        x = self.activation(x)

        if pos == 2:
            x = self.dropout(x)

        x = self.norm_layer(x)
        x = self.conv3(x)
        x = self.activation(x)

        if pos == 3:
            x = self.dropout(x)
        return x


class DSCBlock(Module):

    def __init__(self, in_, out_, pool=(2, 1), activation=ReLU, dropout=0.4):
        super(DSCBlock, self).__init__()

        self.activation = activation()
        self.conv1 = DepthSepConv2D(in_, out_, kernel_size=(3, 3))
        self.conv2 = DepthSepConv2D(out_, out_, kernel_size=(3, 3))
        self.conv3 = DepthSepConv2D(out_, out_, kernel_size=(3, 3), padding=(1, 1), stride=pool)
        self.norm_layer = InstanceNorm2d(out_, eps=0.001, momentum=0.99, track_running_stats=False)
        self.dropout = MixDropout(dropout_proba=dropout, dropout2d_proba=dropout/2)

    def forward(self, x):
        pos = random.randint(1, 3)
        x = self.conv1(x)
        x = self.activation(x)

        if pos == 1:
            x = self.dropout(x)

        x = self.conv2(x)
        x = self.activation(x)

        if pos == 2:
            x = self.dropout(x)

        x = self.norm_layer(x)
        x = self.conv3(x)

        if pos == 3:
            x = self.dropout(x)
        return x


import torch
from torch.utils.data import Dataset
from transformers import AutoTokenizer, TrOCRProcessor
from PIL import Image

class CustomDataset(Dataset):
    def __init__(self, anot, config):
        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained("vinai/bartpho-syllable-base")
        self.processor = TrOCRProcessor.from_pretrained("microsoft/trocr-small-handwritten")
        self.anot = anot
        self.max_seq_len = config.max_seq_len

    def __len__(self):
        return len(self.anot)

    def __getitem__(self, idx):
        sample = self.anot.iloc[idx]
        fn = sample.filename

        input_ids = self.tokenizer(str(sample.label), return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_seq_len)

        img_path = "augmented_images/" + fn
        if fn[:4] == "wild":
            img_path = "WildLine/" + fn
        if fn[:5] == "digit":
            img_path = "digits/" + fn
        if fn[:6] == "single":
            img_path = "single_digit/" + fn

        image = Image.open(img_path).convert("RGB")
        pixel_values = self.processor(image, return_tensors="pt").pixel_values

        return {
            "pixel_values":    pixel_values[0],
            "input_ids":   input_ids["input_ids"][0],
            "att_mask":     input_ids["attention_mask"][0]
        }


from models import *
from utils import *
from dataset import *
from torch.utils.data import DataLoader
import pandas as pd
from tqdm import tqdm
from evaluate import load
import wandb
import math

device="cuda:1"

if __name__ == '__main__':
    #define model
    torch.set_default_dtype(torch.float32)
    torch.autograd.set_detect_anomaly(True)
    model = VGGTransformer(config)
    model.load_state_dict(torch.load("source/weights/cp_vgg_transformer_augmented_data.pt"))
    model.to(device)

    #data loader
    anot = pd.read_csv("finetune_anot.csv").sample(frac=1, random_state=0).reset_index(drop=True)

    traindataset = CustomDataset(anot[:-10000], config)
    trainloader = DataLoader(dataset=traindataset, batch_size=24, shuffle=True, generator=torch.Generator(device="cpu"), num_workers=8)
    valdataset = CustomDataset(pd.read_csv("test_anot.csv")[500:], config)
    valloader = DataLoader(dataset=valdataset, batch_size=196, shuffle=False, generator=torch.Generator(device="cpu"), num_workers=8)

    #optimizer
    param1, param2 = param_to_update(model)
    optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00001},
                                {'params': param2, 'lr': 0.00001}], weight_decay=1e-4)

    #loss function
    lsLoss = LabelSmoothingLoss(233, padding_idx=0, smoothing=0.1)

    #utils
    cer = load("cer")
    tkz = Tokenizer(config)

    #number of epochs
    n_epoch = 10
    #checkpoint
    best_cer = 1

    # wandb init
    # wandb.login(key = '')
    # wandb.login(key = '')
    # wandb.init(
    #   project="VNHTR",
    #   name=f"vgg_transformers_finetune_v3_5",
    #   config={
    #       "learning_rate": 0.0001,
    #       "architecture": "v_t",
    #       "dataset": "gen",
    #       "epochs": 10,
    #   }
    # )

    max_len = {1: 8, 2: 16, 3: 21, 4: 26, 5: 33, 6: 38, 7: 43, 8: 49, 9: 54, 10: 64, 11: 60, 12: 63, 13: 66, 14: 66, 15: 66, 16: 70, 17: 80, 18: 90, 19:108, 20: 106, 21: 114}

    print("Start .....")
    for epoch in range(n_epoch):
        print(f"epoch {epoch} ----------------------------")
        if epoch < 1:
            optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00005},
                            {'params': param2, 'lr': 0.0001}], weight_decay=1e-5)
        # elif epoch < 3:
        #     optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00005},
        #                     {'params': param2, 'lr': 0.00005}], weight_decay=1e-5)
        else:
            optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00001},
                            {'params': param2, 'lr': 0.00001}], weight_decay=1e-5)
        model.train()
        train_loss = 0
        # torch.cuda.empty_cache()
        for batch in tqdm(trainloader):
            img = batch["image"].to(device)
            in_tokens = batch["tokens"][:,:-1].to(device)
            padding_mask = batch["padding_mask"][:,:-1].to(device)
            max_width = batch["width"].max().item()
            max_token = min(max_len[min(math.ceil(max_width/config.im_height),21)] + 5, config.max_seq_len)

            optimizer.zero_grad()

            # forward
            outs = model(img[:, :, :, :max_width], in_tokens[:, :max_token], padding_mask[:, :max_token])

            # ground truth
            tokens = batch["tokens"][:,1:].to(device)

            loss = lsLoss(outs, tokens[:, :max_token])
            
            loss.backward()
            
            torch.nn.utils.clip_grad_norm_(model.parameters(), 4)
            
            optimizer.step()

            train_loss += loss.item()

        else:
            print(f"train loss: {train_loss/len(trainloader)}")
            # torch.save(model.state_dict(), "source/weights/last_vgg_transformer.pt")
            if True:
                with torch.no_grad():
                    model.eval()
                    predictions, references = [], []
                    torch.cuda.empty_cache()
                    for batch in tqdm(valloader):
                        img = batch["image"].to(device)
                        tokens = batch["tokens"][:,1:].to(device)
                        max_width = batch["width"].max().item()

                        t_preds = infer_transformer_base(img[:, :, :, :max_width], model.to(device))

                        chars_true = [tkz.reverse_tokens(token) for token in tokens]
                        chars_pred = [tkz.reverse_tokens(t_pred) for t_pred in t_preds]

                        predictions += chars_pred
                        references += chars_true

                    cer_score = cer.compute(predictions=predictions, references=references, concatenate_texts=True)
                    # wandb.log({"train_loss": train_loss/len(trainloader), "cer": cer_score})
                    if cer_score < best_cer:
                        best_cer = cer_score
                        torch.save(model.state_dict(), "source/weights/cp_vgg_transformer_finetune_v4.pt")

                    print(f'Character Error Rate: {cer_score}')
    
    # wandb.finish()


import torch
import numpy as np
from torch.nn import functional as F
import os
import tempfile
from PIL import Image, ImageOps
import gdown
from transformers import AutoTokenizer, TrOCRProcessor

class Tokenizer():
    def __init__(self):
        vocab = ['a', 'A', 'à', 'À', 'ả', 'Ả', 'ã', 'Ã', 'á', 'Á', 'ạ', 'Ạ', 'ă', 'Ă', 'ằ', 'Ằ', 'ẳ', 'Ẳ', 'ẵ', 'Ẵ', 'ắ', 'Ắ', 'ặ', 'Ặ', 'â', 'Â', 'ầ', 'Ầ', 'ẩ', 'Ẩ', 'ẫ', 'Ẫ', 'ấ', 'Ấ', 'ậ', 'Ậ', 'b', 'B', 'c', 'C', 'd', 'D', 'đ', 'Đ', 'e', 'E', 'è', 'È', 'ẻ', 'Ẻ', 'ẽ', 'Ẽ', 'é', 'É', 'ẹ', 'Ẹ', 'ê', 'Ê', 'ề', 'Ề', 'ể', 'Ể', 'ễ', 'Ễ', 'ế', 'Ế', 'ệ', 'Ệ', 'f', 'F', 'g', 'G', 'h', 'H', 'i', 'I', 'ì', 'Ì', 'ỉ', 'Ỉ', 'ĩ', 'Ĩ', 'í', 'Í', 'ị', 'Ị', 'j', 'J', 'k', 'K', 'l', 'L', 'm', 'M', 'n', 'N', 'o', 'O', 'ò', 'Ò', 'ỏ', 'Ỏ', 'õ', 'Õ', 'ó', 'Ó', 'ọ', 'Ọ', 'ô', 'Ô', 'ồ', 'Ồ', 'ổ', 'Ổ', 'ỗ', 'Ỗ', 'ố', 'Ố', 'ộ', 'Ộ', 'ơ', 'Ơ', 'ờ', 'Ờ', 'ở', 'Ở', 'ỡ', 'Ỡ', 'ớ', 'Ớ', 'ợ', 'Ợ', 'p', 'P', 'q', 'Q', 'r', 'R', 's', 'S', 't', 'T', 'u', 'U', 'ù', 'Ù', 'ủ', 'Ủ', 'ũ', 'Ũ', 'ú', 'Ú', 'ụ', 'Ụ', 'ư', 'Ư', 'ừ', 'Ừ', 'ử', 'Ử', 'ữ', 'Ữ', 'ứ', 'Ứ', 'ự', 'Ự', 'v', 'V', 'w', 'W', 'x', 'X', 'y', 'Y', 'ỳ', 'Ỳ', 'ỷ', 'Ỷ', 'ỹ', 'Ỹ', 'ý', 'Ý', 'ỵ', 'Ỵ', 'z', 'Z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`', '{', '|', '}', '~', ' ']
        vocab = ["pad", "<sos>", "<eos>", "<unk>"] + vocab

        self.c_vocab = dict(zip(vocab, np.arange(len(vocab))))
        self.reverse_vocab = list(self.c_vocab.items())
        self.max_seq_len = 128

    def tokenize(self, seq):
        res = []
        for c in ['<sos>'] + seq + ['<eos>']:
            try:
                res.append(self.c_vocab[c])
            except:
                res.append(self.c_vocab["<unk>"])
        mask = [1]*len(res)
        if len(res) < self.max_seq_len:
            n_pad = self.max_seq_len-len(res)+1
            res += [self.c_vocab["pad"]]*n_pad
            mask += [0]*n_pad
        elif len(res) > self.max_seq_len:
            s_pad = self.max_seq_len
            res = res[:s_pad] + [self.c_vocab["<eos>"]]
            mask = mask[:s_pad+1]
        else:
            res += [self.c_vocab["pad"]]
            mask += [0]
        return np.array(res), np.logical_not(mask)

    def reverse_tokens(self, tokens):
        res = []
        for token in tokens:
            if token in [self.c_vocab["<unk>"], self.c_vocab["pad"], self.c_vocab["<sos>"]]:
                continue
            if token == self.c_vocab["<eos>"]:
                break
            res.append(self.reverse_vocab[token][0])
            
        return ''.join(res)

    def reverse_tokens_special(self, tokens):
        res = []
        for token in tokens:
            res.append(self.reverse_vocab[token][0])
            # if token == self.c_vocab["<eos>"]:
            #     break
            
        return res

class VGGTransformer():
    def __init__(self, device="cuda:2"):
        encoder_path, decoder_path = self.download_weights()
        self.tkz = Tokenizer()
        self.encoder = torch.jit.load(encoder_path, map_location=device)#.to(device)
        self.decoder = torch.jit.load(decoder_path, map_location=device)#.to(device)
        self.encoder.eval()
        self.decoder.eval()
        self.device = device
        self.warmup()

    def download_weights(self):
        encoder_path = os.path.join(tempfile.gettempdir(), "vta_encoder.pt")
        decoder_path = os.path.join(tempfile.gettempdir(), "vta_decoder.pt")
        if os.path.exists(encoder_path):
            print(f'VGG Transformer with Rethinking Head encoder weights {encoder_path} exsits. Ignore download!')
        else:
            print(f'Downloading VGG Transformer with Rethinking Head encoder weights {encoder_path} ...')
            if os.system(f'wget -O {encoder_path} https://github.com/nguyenhoanganh2002/vnhtr_weights/blob/main/vta_encoder.pt?raw=true'):
                raise RuntimeError('Download encoder failed!')
            # gdown.download(id="179BRTjPBMQn5vd9Ah6DIklcox30zh4YQ", output=encoder_path, quiet=True)
            # if not os.path.exists(encoder_path):
            #     print("Retrying ...")
            #     if os.system(f'curl -H "Authorization: Bearer ya29.a0AfB_byDhvrZegVfgDHMqnSXRhp763RCQjw7HhwBR-eN3DCTmx-Q6SlsXpd5QagdhICn2zy5Dpp8SVYRWiDuwhH-IemClyvaElQFiOQBYSL7Hxy_ddAEEv6HbuxbzcKKtvRnaXpvSOIJ8ui8g1O93iY0tFHO7hFLGfy9PaCgYKAYASARMSFQHGX2MiAwlU1xpS3ZceA0-121l89w0171" https://www.googleapis.com/drive/v3/files/179BRTjPBMQn5vd9Ah6DIklcox30zh4YQ?alt=media -o {encoder_path}'):
            #         raise RuntimeError('Download encoder failed!')
            
        if os.path.exists(decoder_path):
            print(f'VGG Transformer with Rethinking Head decoder weights {decoder_path} exsits. Ignore download!')
        else:
            print(f'Downloading VGG Transformer with Rethinking Head decoder weights {decoder_path} ...')
            if os.system(f'wget -O {decoder_path} https://github.com/nguyenhoanganh2002/vnhtr_weights/blob/main/vta_decoder.pt?raw=true'):
                raise RuntimeError('Download decoder failed!')
            # gdown.download(id="1dNJrjBF-FcjQgzckr4CKJyFdaLRuse6q", output=decoder_path, quiet=True)
            # if not os.path.exists(decoder_path):
            #     print("Retrying ...")
            #     if os.system(f'curl -H "Authorization: Bearer ya29.a0AfB_byDhvrZegVfgDHMqnSXRhp763RCQjw7HhwBR-eN3DCTmx-Q6SlsXpd5QagdhICn2zy5Dpp8SVYRWiDuwhH-IemClyvaElQFiOQBYSL7Hxy_ddAEEv6HbuxbzcKKtvRnaXpvSOIJ8ui8g1O93iY0tFHO7hFLGfy9PaCgYKAYASARMSFQHGX2MiAwlU1xpS3ZceA0-121l89w0171" https://www.googleapis.com/drive/v3/files/1dNJrjBF-FcjQgzckr4CKJyFdaLRuse6q?alt=media -o {decoder_path}'):
            #         raise RuntimeError('Download decoder failed!')
        return encoder_path, decoder_path

    def warmup(self):
        print("Warming up ...")
        img = torch.rand(2,3,32,768)
        with torch.no_grad():
            with torch.jit.optimized_execution(True):
                mem = self.encoder(img.to(self.device))
                translated_sentence = [[1]*len(img)]
                max_length = 0

                while max_length <= 128 and not all(np.any(np.asarray(translated_sentence).T==2, axis=1)):

                    tgt_inp = torch.LongTensor(translated_sentence).to(self.device)
                    # print(self.gen_nopeek_mask(tgt_inp.shape[0]).to(self.device).device)
                    output = self.decoder(tgt_inp, mem, self.gen_nopeek_mask(tgt_inp.shape[0]).to(self.device))
                    output = F.softmax(output, dim=-1)
                    output = output.to('cpu')

                    values, indices  = torch.topk(output, 5)
                    
                    indices = indices[:, -1, 0]
                    indices = indices.tolist()

                    translated_sentence.append(indices)   
                    max_length += 1

                    del output

                translated_sentence = np.asarray(translated_sentence).T
    
    def predict(self, imgs, max_seq_length=128, sos_token=1, eos_token=2):
        max_W = 0
        ims = []
        for img in imgs:
            im, W = self.process_input(img)
            ims.append(im)
            if max_W < W:
                max_W = W
        imgs = torch.FloatTensor(np.array(ims))[:,:,:,:max_W]

        with torch.no_grad():
            with torch.jit.optimized_execution(True):
                mem = self.encoder(imgs.to(self.device))
                translated_sentence = [[sos_token]*len(imgs)]
                max_length = 0

                while max_length <= max_seq_length and not all(np.any(np.asarray(translated_sentence).T==eos_token, axis=1)):

                    tgt_inp = torch.LongTensor(translated_sentence).to(self.device)
                    # print(self.gen_nopeek_mask(tgt_inp.shape[0]).to(self.device).device)
                    output = self.decoder(tgt_inp, mem, self.gen_nopeek_mask(tgt_inp.shape[0]).to(self.device))
                    output = F.softmax(output, dim=-1)
                    output = output.to('cpu')

                    values, indices  = torch.topk(output, 5)
                    
                    indices = indices[:, -1, 0]
                    indices = indices.tolist()

                    translated_sentence.append(indices)   
                    max_length += 1

                    del output

                translated_sentence = np.asarray(translated_sentence).T
            
                return [self.tkz.reverse_tokens(tokens) for tokens in translated_sentence]
    
    def gen_nopeek_mask(self, length):
        mask = (torch.triu(torch.ones(length, length)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask

    def process_input(self, image):
        img = image.convert('RGB')
        image_height = 32

        w, h = img.size
        new_w = int(image_height*w/h)

        img = img.resize((new_w, image_height), Image.LANCZOS)
        img = ImageOps.expand(img, border=(0, 0, 768 - new_w, 0), fill='white')
        img = np.asarray(img).transpose(2,0, 1)
        img = img/255
        return img, new_w
    
class TrOCR:
    def __init__(self, device="cuda:0"):
        encoder_path, decoder_path = self.download_weights()
        self.tokenizer = AutoTokenizer.from_pretrained("vinai/bartpho-syllable-base")
        self.processor = TrOCRProcessor.from_pretrained("microsoft/trocr-small-handwritten")
        self.encoder = torch.jit.load(encoder_path, map_location=device)#.to(device)
        self.decoder = torch.jit.load(decoder_path, map_location=device)#.to(device)
        self.encoder.eval()
        self.decoder.eval()
        self.device = device
        self.warmup()

    def download_weights(self):
        encoder_path = os.path.join(tempfile.gettempdir(), "tra_encoder.pt")
        decoder_path = os.path.join(tempfile.gettempdir(), "tra_decoder.pt")
        if os.path.exists(encoder_path):
            print(f'TrOCR with Rethinking Head encoder weights {encoder_path} exsits. Ignore download!')
        else:
            print(f'Downloading TrOCR with Rethinking Head encoder weights {encoder_path} ...')
            if os.system(f'wget -O {encoder_path} https://github.com/nguyenhoanganh2002/vnhtr_weights/blob/main/tra_encoder.pt?raw=true'):
                raise RuntimeError('Download encoder failed!')
            # gdown.download(id="1HnFESJHpvefkeqEovMZrOBgkYSRw2owK", output=encoder_path, quiet=True)
            # if not os.path.exists(encoder_path):
            #     print("Retrying ...")
            #     if os.system(f'curl -H "Authorization: Bearer ya29.a0AfB_byDhvrZegVfgDHMqnSXRhp763RCQjw7HhwBR-eN3DCTmx-Q6SlsXpd5QagdhICn2zy5Dpp8SVYRWiDuwhH-IemClyvaElQFiOQBYSL7Hxy_ddAEEv6HbuxbzcKKtvRnaXpvSOIJ8ui8g1O93iY0tFHO7hFLGfy9PaCgYKAYASARMSFQHGX2MiAwlU1xpS3ZceA0-121l89w0171" https://www.googleapis.com/drive/v3/files/1HnFESJHpvefkeqEovMZrOBgkYSRw2owK?alt=media -o {encoder_path}'):
            #         raise RuntimeError('Download encoder failed!')
            
        if os.path.exists(decoder_path):
            print(f'TrOCR with Rethinking Head decoder weights {decoder_path} exsits. Ignore download!')
        else:
            print(f'Downloading TrOCR with Rethinking Head decoder weights {decoder_path} ...')
            if os.system(f'wget -O {decoder_path} https://github.com/nguyenhoanganh2002/vnhtr_weights/blob/main/tra_decoder.pt?raw=true'):
                raise RuntimeError('Download decoder failed!')
            # gdown.download(id="1E4hbRQoqQjLHueMaV3qsxI7rUI0uPgmJ", output=decoder_path, quiet=True)
            # if not os.path.exists(decoder_path):
            #     print("Retrying ...")
            #     if os.system(f'curl -H "Authorization: Bearer ya29.a0AfB_byDhvrZegVfgDHMqnSXRhp763RCQjw7HhwBR-eN3DCTmx-Q6SlsXpd5QagdhICn2zy5Dpp8SVYRWiDuwhH-IemClyvaElQFiOQBYSL7Hxy_ddAEEv6HbuxbzcKKtvRnaXpvSOIJ8ui8g1O93iY0tFHO7hFLGfy9PaCgYKAYASARMSFQHGX2MiAwlU1xpS3ZceA0-121l89w0171" https://www.googleapis.com/drive/v3/files/1E4hbRQoqQjLHueMaV3qsxI7rUI0uPgmJ?alt=media -o {decoder_path}'):
            #         raise RuntimeError('Download decoder failed!')
        return encoder_path, decoder_path

    def warmup(self):
        print("Warming up ...")
        img = torch.randint(0, 256, (2,3,512,1024))
        pixel_values = self.processor(img, return_tensors="pt").pixel_values.to(self.device)
        with torch.no_grad():
            with torch.jit.optimized_execution(True):
                encoder_output = self.encoder(pixel_values)
    
                start_ids = torch.LongTensor([[0]]*pixel_values.shape[0]).to(self.device)

                max_length = 0

                while max_length <= 35 and not all(start_ids[:, -1] == 2):
                    output = self.decoder(start_ids, encoder_output)
                    output = F.softmax(output[:,-1,:], dim=-1)

                    start_ids = torch.cat([start_ids, output.argmax(dim=-1).unsqueeze(1)], dim=-1)

                    max_length += 1
    
    def predict(self, imgs, max_seq_length=35, sos_token=0, eos_token=2):
        pixel_values = []
        for img in imgs:
            im = self.processor(img, return_tensors="pt").pixel_values
            pixel_values.append(im)
        pixel_values = torch.FloatTensor(np.array(pixel_values))
        with torch.no_grad():
            with torch.jit.optimized_execution(True):
                encoder_output = self.encoder(pixel_values.to(self.device))
    
                start_ids = torch.LongTensor([[sos_token]]*pixel_values.shape[0]).to(self.device)

                max_length = 0

                while max_length <= max_seq_length and not all(start_ids[:, -1] == eos_token):
                    output = self.decoder(start_ids, encoder_output, self.gen_nopeek_mask(start_ids.shape[1]).to(self.device))
                    output = F.softmax(output[:,-1,:], dim=-1)

                    start_ids = torch.cat([start_ids, output.argmax(dim=-1).unsqueeze(1)], dim=-1)

                    max_length += 1
            
            return self.tokenizer.batch_decode(start_ids, skip_special_tokens=True)
    
    def gen_nopeek_mask(self, length):
        mask = (torch.triu(torch.ones(length, length)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask


from model import *
from utils import *
from dataset import *
from torch.utils.data import DataLoader
import pandas as pd
from tqdm import tqdm
from evaluate import load
import wandb
from config import *


if __name__ == '__main__':
    #define model
    torch.set_default_dtype(torch.float32)
    torch.autograd.set_detect_anomaly(True)
    model = VNTrOCR()
    model.load_state_dict(torch.load("source/weights/cp_vision_encoder_decoder_augmented_data.pt"))
    model.to(device)

    #data loader
    anot = pd.read_csv("concated_anot.csv").sample(frac=1, random_state=0).reset_index(drop=True)

    traindataset = CustomDataset(anot[:-10000], config)
    trainloader = DataLoader(dataset=traindataset, batch_size=16, shuffle=True, generator=torch.Generator(device="cpu"), num_workers=8)
    valdataset = CustomDataset(anot[-10000:], config)
    valloader = DataLoader(dataset=valdataset, batch_size=16, shuffle=False, generator=torch.Generator(device="cpu"), num_workers=8)

    #optimizer
    param1, param2 = param_to_update(model)
    optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.0001},
                                {'params': param2, 'lr': 0.0001}], weight_decay=1e-5)

    #loss function
    # lsLoss = LabelSmoothingLoss(233, padding_idx=0, smoothing=0.1)

    #utils
    cer = load("cer")

    #number of epochs
    n_epoch = 15
    #checkpoint
    best_cer = 1

    # wandb init
    # wandb.login(key = '')
    # wandb.login(key = '')
    # wandb.init(
    #   project="VN_HTR",
    #   name=f"vision_encoder_decoder_augmented_data",
    #   config={
    #       "learning_rate": 0.0001,
    #       "architecture": "v_t",
    #       "dataset": "gen",
    #       "epochs": 10,
    #   }
    # )


    print("Start .....")
    for epoch in range(n_epoch):
        print(f"epoch {epoch} ----------------------------")
        if epoch < 1:
            optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00005},
                            {'params': param2, 'lr': 0.00005}], weight_decay=1e-5)
        elif epoch < 3:
            optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.00001},
                            {'params': param2, 'lr': 0.00001}], weight_decay=1e-5)
        else:
            optimizer = torch.optim.Adam([{'params': param1, 'lr': 0.000005},
                            {'params': param2, 'lr': 0.000005}], weight_decay=1e-5)
        model.train()
        train_loss = 0
        torch.cuda.empty_cache()
        for batch in tqdm(trainloader):
            pixel_values = batch["pixel_values"].to(device)
            input_ids = batch["input_ids"].to(device)
            att_mask = batch["att_mask"].to(device)

            optimizer.zero_grad()

            # forward
            loss = model(pixel_values, input_ids, att_mask)
            
            loss.backward()
            
            torch.nn.utils.clip_grad_norm_(model.parameters(), 4)
            
            optimizer.step()

            train_loss += loss.item()

        else:
            print(f"train loss: {train_loss/len(trainloader)}")
            # torch.save(model.state_dict(), "source/weights/last_vgg_transformer.pt")
            if True:
                with torch.no_grad():
                    model.eval()
                    predictions, references = [], []
                    torch.cuda.empty_cache()
                    for batch in tqdm(valloader):
                        pixel_values = batch["pixel_values"].to(device)
                        input_ids = batch["input_ids"].to(device)

                        preds = infer(pixel_values, model)

                        predictions += valdataset.tokenizer.batch_decode(preds, skip_special_tokens=True)
                        references += valdataset.tokenizer.batch_decode(input_ids, skip_special_tokens=True)
                    for i in range(len(references)):
                        if references[i] == "":
                            references[i] = " "
                    cer_score = 1
                    try:
                        cer_score = cer.compute(predictions=predictions, references=references, concatenate_texts=True)
                    except:
                        print("error when caculate cer score")
                    # wandb.log({"train_loss": train_loss/len(trainloader), "cer": cer_score})
                    if cer_score < best_cer:
                        best_cer = cer_score
                        torch.save(model.state_dict(), "source/weights/cp_finetune.pt")

                    print(f'Character Error Rate: {cer_score}')
    
    # wandb.finish()


import numpy as np
import math
import torch
from torch.utils.data import Dataset
from PIL import Image, ImageOps
from config import Config, device

config = Config()

class Tokenizer():
    def __init__(self, config):
        vocab = ['a', 'A', 'à', 'À', 'ả', 'Ả', 'ã', 'Ã', 'á', 'Á', 'ạ', 'Ạ', 'ă', 'Ă', 'ằ', 'Ằ', 'ẳ', 'Ẳ', 'ẵ', 'Ẵ', 'ắ', 'Ắ', 'ặ', 'Ặ', 'â', 'Â', 'ầ', 'Ầ', 'ẩ', 'Ẩ', 'ẫ', 'Ẫ', 'ấ', 'Ấ', 'ậ', 'Ậ', 'b', 'B', 'c', 'C', 'd', 'D', 'đ', 'Đ', 'e', 'E', 'è', 'È', 'ẻ', 'Ẻ', 'ẽ', 'Ẽ', 'é', 'É', 'ẹ', 'Ẹ', 'ê', 'Ê', 'ề', 'Ề', 'ể', 'Ể', 'ễ', 'Ễ', 'ế', 'Ế', 'ệ', 'Ệ', 'f', 'F', 'g', 'G', 'h', 'H', 'i', 'I', 'ì', 'Ì', 'ỉ', 'Ỉ', 'ĩ', 'Ĩ', 'í', 'Í', 'ị', 'Ị', 'j', 'J', 'k', 'K', 'l', 'L', 'm', 'M', 'n', 'N', 'o', 'O', 'ò', 'Ò', 'ỏ', 'Ỏ', 'õ', 'Õ', 'ó', 'Ó', 'ọ', 'Ọ', 'ô', 'Ô', 'ồ', 'Ồ', 'ổ', 'Ổ', 'ỗ', 'Ỗ', 'ố', 'Ố', 'ộ', 'Ộ', 'ơ', 'Ơ', 'ờ', 'Ờ', 'ở', 'Ở', 'ỡ', 'Ỡ', 'ớ', 'Ớ', 'ợ', 'Ợ', 'p', 'P', 'q', 'Q', 'r', 'R', 's', 'S', 't', 'T', 'u', 'U', 'ù', 'Ù', 'ủ', 'Ủ', 'ũ', 'Ũ', 'ú', 'Ú', 'ụ', 'Ụ', 'ư', 'Ư', 'ừ', 'Ừ', 'ử', 'Ử', 'ữ', 'Ữ', 'ứ', 'Ứ', 'ự', 'Ự', 'v', 'V', 'w', 'W', 'x', 'X', 'y', 'Y', 'ỳ', 'Ỳ', 'ỷ', 'Ỷ', 'ỹ', 'Ỹ', 'ý', 'Ý', 'ỵ', 'Ỵ', 'z', 'Z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`', '{', '|', '}', '~', ' ']
        vocab = ["pad", "<sos>", "<eos>", "<unk>"] + vocab

        self.c_vocab = dict(zip(vocab, np.arange(len(vocab))))
        self.reverse_vocab = list(self.c_vocab.items())
        self.max_seq_len = config.max_seq_len

    def tokenize(self, seq):
        res = []
        for c in ['<sos>'] + seq + ['<eos>']:
            try:
                res.append(self.c_vocab[c])
            except:
                res.append(self.c_vocab["<unk>"])
        mask = [1]*len(res)
        if len(res) < self.max_seq_len:
            n_pad = self.max_seq_len-len(res)+1
            res += [self.c_vocab["pad"]]*n_pad
            mask += [0]*n_pad
        elif len(res) > self.max_seq_len:
            s_pad = self.max_seq_len
            res = res[:s_pad] + [self.c_vocab["<eos>"]]
            mask = mask[:s_pad+1]
        else:
            res += [self.c_vocab["pad"]]
            mask += [0]
        return np.array(res), np.logical_not(mask)

    def reverse_tokens(self, tokens):
        res = []
        for token in tokens:
            if token in [self.c_vocab["<unk>"], self.c_vocab["pad"], self.c_vocab["<sos>"]]:
                continue
            if token == self.c_vocab["<eos>"]:
                break
            res.append(self.reverse_vocab[token][0])
            
        return ''.join(res)

    def reverse_tokens_special(self, tokens):
        res = []
        for token in tokens:
            res.append(self.reverse_vocab[token][0])
            # if token == self.c_vocab["<eos>"]:
            #     break
            
        return res

def process_input(image):
    img = image.convert('RGB')
    image_height = config.im_height

    w, h = img.size
    new_w = int(image_height*w/h)

    img = img.resize((new_w, image_height), Image.LANCZOS)
    img = ImageOps.expand(img, border=(0, 0, config.im_width - new_w, 0), fill='white')
    img = np.asarray(img).transpose(2,0, 1)
    img = img/255
    return img, new_w

class CustomDataset(Dataset):
    def __init__(self, anot, config):
        super().__init__()
        self.gen_source = "images_gen/"
        self.grey_source = "grey/"
        self.word_source = "grey_word/"
        self.tokenizer = Tokenizer(config)
        self.anot = anot

    def __len__(self):
        return len(self.anot)

    def __getitem__(self, idx):
        sample = self.anot.iloc[idx]
        fn = sample.filename
        # if fn[0] == 'g':
        #     img_path = self.gen_source + fn
        # elif fn[0] == 'w':
        #     img_path = self.word_source + fn
        # else:
        #     img_path = self.grey_source + fn
        
        chars = list(sample.label)
        tokens, mask = self.tokenizer.tokenize(chars)

        img_path = "augmented_images/" + fn
        if fn[:4] == "wild":
            img_path = "WildLine/" + fn
        if fn[:5] == "digit":
            img_path = "digits/" + fn
        if fn[:6] == "single":
            img_path = "single_digit/" + fn

        image = Image.open(img_path)
        image, width = process_input(image)

        return {
            "image":    torch.FloatTensor(image).to("cpu"),
            "tokens":   torch.LongTensor(tokens).to("cpu"),
            "padding_mask":     torch.FloatTensor(mask).to("cpu"),
            "width": torch.LongTensor([width]).to("cpu")
        }